`timescale 1 ps / 1 ps

module a_tb ();

// Constants
parameter	REF_CLK_PERIOD 			= 10000;		//100MHz
parameter	MGMT_CLK_PERIOD 		= 10000;		//100MHz
parameter	counter_width_0 = 32;
parameter	COUNT_VALUE = 9;	// Specify the number of slip needed
parameter	PMA_WIDTH = 256;		// Specify your PMA width settings
parameter	RECONFIG_ADD_WIDTH = 11;

// General purpose registers
reg clock_tb;
reg mgmt_blk_clock_tb;
reg reset_tb;
reg [PMA_WIDTH - 1:0] tx_parallel_data_tb;

reg [39:0] tx_bitslip_tb;
reg rx_set_locktodata_tb;
reg rx_set_locktoref_tb;
reg rx_seriallpbken_tb;
reg tx_std_polinv_tb;

reg pll_select_tb;

// Bit slip functions
reg		[counter_width_0 - 1:0]	counter_0;
reg		start_count;
reg		[counter_width_0-1:0]	count_val;
reg [1:0] rx_bitslip_tb;


// Reconfig management input
reg reconfig_mgmt_read_tb;
reg reconfig_mgmt_write_tb;
reg [RECONFIG_ADD_WIDTH - 1:0] reconfig_mgmt_address_tb;
reg [31:0] reconfig_mgmt_writedata_tb;

reg stream_req_tb;
reg cfg_sel_tb;

// PRBS verifier
reg rx_10g_prbs_err_clr_tb;
wire rx_10g_prbs_done_tb;
wire rx_10g_prbs_err_tb;


// Reconfig management
wire reconfig_busy_tb;
wire reconfig_mgmt_waitrequest_tb;
wire [31:0] reconfig_mgmt_readdata_tb;

wire	count_done;

wire	stream_busy_tb;

// Wires
wire [1:0] tx_ready_tb;
wire [1:0] rx_ready_tb;
wire pll_locked_tb;
wire [1:0] tx_serial_data_tb;
wire [1:0] rx_serial_data_tb;

wire [1:0] rx_is_lockedtoref_tb;
wire [1:0] rx_is_lockedtodata_tb;
wire tx_clkout_tb;
wire rx_clkout_tb;
wire [PMA_WIDTH - 1:0] rx_parallel_data_tb;
wire rx_std_signaldetect_tb;

wire [1:0] tx_serial_clk0_tb;
wire [1:0] tx_serial_clkb_tb;
wire tx_serial_clk1g_tb;
wire tx_serial_clk2g_tb;

// External loopback TX -> RX
assign		rx_serial_data_tb = tx_serial_data_tb;

// Connect the pll serial clock
assign tx_serial_clk0_tb[0] = tx_serial_clk2g_tb;
assign tx_serial_clk0_tb[1] = tx_serial_clk2g_tb;
assign tx_serial_clkb_tb[0] = tx_serial_clk1g_tb;
assign tx_serial_clkb_tb[1] = tx_serial_clk1g_tb;

// Start the test

initial
begin

	tx_parallel_data_tb = {128'h0BC, 128'h0BC};

       end

initial
begin
	//tx_parallel_data_tb = 32'hBCBCBCBC;	
	start_count = 1'b0;
	count_val = 32'h00000000;
	reset_tb 		= 1'b0;
	pll_select_tb = 1'b0;	// Only one TX PLL available in the Native PHy
	stream_req_tb = 1'b0;
	cfg_sel_tb	= 1'b0;


	// rx_set_locktoref_tb & rx_set_locktodata_tb = 0 -> CDR auto lock
	// mode
	rx_set_locktoref_tb	= 1'b0;
	rx_set_locktodata_tb	= 1'b0;
	rx_seriallpbken_tb	= 1'b0;
	rx_bitslip_tb		= 1'b0;
//	tx_bitslip_tb		= 40'h4;
	tx_std_polinv_tb	= 1'b0;
	reconfig_mgmt_read_tb	= 1'b0;
	reconfig_mgmt_write_tb	= 1'b0;
	reconfig_mgmt_address_tb	= 7'b0;
	reconfig_mgmt_writedata_tb	= 32'b0;
	rx_10g_prbs_err_clr_tb  = 1'b0;


	#50000
	reset_tb = 1'b1;
	#100000
	reset_tb = 1'b0;


	// Test read the PLL mux info for switching
	// Request access to AVMM interface of for PLL

	// Switch to TX PLL 1 for 1Gbps
	// Switch TX PLL for CH0, take note of the reconfig_address bit[10]
	#35000000
	write32 (11'h000, 32'h2);
	#2000000
	read32 (11'h117);

	// Logical PLL encoding has been pre-determined from simulation
	#2000000
	write32 (11'h111, 32'h80);
	#2000000
	write32 (11'h3, 32'h0);

	// Switch TX PLL for CH1, take note of the reconfig_address bit[10]
	#2000000
	write32 (11'h400, 32'h2);
	#2000000
	read32 (11'h517);

	// Logical PLL encoding has been pre-determined from simulation
	#2000000
	write32 (11'h511, 32'h80);
	#2000000
	write32 (11'h403, 32'h0);

	//request access to AVMM interface of CH0
	//after request granted, reconfig_waitrequest will = 0
	//This is a MUST or else the subsequent commands to AVMM will not be
	//entertained
	#10000000
	write32 (11'h0, 32'h2);

	// after reconfig_waitrequest = 0 then only can access


	// Embedded streamer register offset info at 0x340
	// bit [7] = cfg_load, bit [6] = bcast_en, bit [2:0] cfg_sel
	//
	// For CH0, kick start the streamer and select profile 0
	// bit [10] = 0, select CH0
	#2000000
	write32 (11'h340, 32'h80);

	
	// Request access to AVMM interface of CH1
	#20000000
	write32 (11'h400, 32'h2);

	// For CH1, kick start the streamer and select profile 0
	// bit [10] = 1, select CH1
	#2000000
	write32 (11'h740, 32'h80);


	// Request access to AVMM interface
	#20000000
	write32 (11'h0, 32'h2);

	// Switch both CH0 and CH1 back to profile 1
	// Kick start the streamer and select profile 1
	// Turn on bcast_en
	#2000000
	write32 (11'h340, 32'hC1);

	// Switch back to TX PLL 0 for 2Gbps
	// Switch TX PLL for CH0, take note of the reconfig_address bit[10]
	#35000000
	write32 (11'h000, 32'h2);
	#2000000
	read32 (11'h117);

	// Logical PLL encoding has been pre-determined from simulation
	#2000000
	write32 (11'h111, 32'hc2);
	#2000000
	write32 (11'h3, 32'h0);

	// Switch TX PLL for CH1, take note of the reconfig_address bit[10]
	#2000000
	write32 (11'h400, 32'h2);
	#2000000
	read32 (11'h517);

	// Logical PLL encoding has been pre-determined from simulation
	#2000000
	write32 (11'h511, 32'hc2);
	#2000000
	write32 (11'h403, 32'h0);

end


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//  Clocks

//reference clock - 125MHz
initial
begin
  clock_tb = 1;
  forever
    #(REF_CLK_PERIOD/2) clock_tb = !clock_tb;
end			

// MGMT reference clock - 125MHz
initial
begin
  mgmt_blk_clock_tb = 1;
  forever
    #(MGMT_CLK_PERIOD/2) mgmt_blk_clock_tb = !mgmt_blk_clock_tb;
end


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//
// #################### multiple bit slip function ################################



assign	count_done = (counter_0 == count_val)? 1'b1 : 1'b0; 

always @(posedge clock_tb or posedge reset_tb)
	if (reset_tb)
		counter_0 = {counter_width_0{1'b0}};
	else if (count_done)
		begin
			counter_0 = {counter_width_0{1'b0}};
			start_count = 1'b0;
		end
	else if (start_count)
		begin
			// Applying bit slips
			#100000
			@(negedge clock_tb) begin
				rx_bitslip_tb			= 1'b1;
			end
			
			#100000
			@(negedge clock_tb) begin
				rx_bitslip_tb			= 1'b0;
			end
					
			counter_0 = counter_0 + 1'b1;
		end

// ################################################################

a 	DUT	(
	.clock			(mgmt_blk_clock_tb),
	.pll_refclk0			(clock_tb),
	.rx_serial_data			(rx_serial_data_tb),
	.tx_parallel_data		(tx_parallel_data_tb),
	.tx_ready			(tx_ready_tb),
	.rx_ready			(rx_ready_tb),
	.rx_bitslip		(rx_bitslip_tb),
	.tx_serial_clk0		(tx_serial_clk0_tb),
	.tx_serial_clkb		(tx_serial_clkb_tb),
	.tx_serial_clk1g		(tx_serial_clk1g_tb),
	.tx_serial_clk2g		(tx_serial_clk2g_tb),
	.tx_serial_data			(tx_serial_data_tb),
	.rx_is_lockedtoref			(rx_is_lockedtoref_tb),
	.rx_parallel_data		(rx_parallel_data_tb),
	.reset				(reset_tb),
	.pll_select			(pll_select_tb),
	.reconfig_read		(reconfig_mgmt_read_tb),
	.reconfig_write		(reconfig_mgmt_write_tb),
	.reconfig_address		(reconfig_mgmt_address_tb),
	.reconfig_writedata	(reconfig_mgmt_writedata_tb),
	.reconfig_waitrequest	(reconfig_mgmt_waitrequest_tb),
	.reconfig_readdata		(reconfig_mgmt_readdata_tb)
);

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Task to perform write
task write32;
input [RECONFIG_ADD_WIDTH - 1:0] reconfig_mgmt_address;
input [31:0] reconfig_mgmt_writedata;

begin
	@(posedge mgmt_blk_clock_tb) begin

		// reconfig address for logical channel
		reconfig_mgmt_address_tb = reconfig_mgmt_address;
		reconfig_mgmt_writedata_tb = reconfig_mgmt_writedata;
	end

	//#1000
	@(posedge mgmt_blk_clock_tb) begin

		// Write to register
		reconfig_mgmt_write_tb = 1'b1;
	end

	// Reset the reconfig write signal
	@(posedge mgmt_blk_clock_tb) begin

		reconfig_mgmt_write_tb = 1'b0;
	end
	
end
endtask

// Task to perform read
task read32;
input [RECONFIG_ADD_WIDTH - 1:0] reconfig_mgmt_address;

begin
@(posedge mgmt_blk_clock_tb) begin

		// reconfig address of busy register
		reconfig_mgmt_address_tb = reconfig_mgmt_address;
	end

	//#1000
	@(posedge mgmt_blk_clock_tb) begin

		// read register
		reconfig_mgmt_read_tb = 1'b1;
	end

	// Reset the reconfig read signal
	@(posedge mgmt_blk_clock_tb) begin

		reconfig_mgmt_read_tb = 1'b0;
	end
end
	
endtask

endmodule
